import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

import "K1R4D3L.KillTracker.Utils";
import "K1R4D3L.KillTracker.UI";

-- identify and set locale
--    beware, if the locale is changed for a character after data is recorded
--    kill counts will not increment correctly if the mobs have different
--    names across the two languages 
_G.locale = "en";

if Turbine.Shell.IsCommand("hilfe") then
	_G.locale = "de";
elseif Turbine.Shell.IsCommand("aide") then
	_G.locale = "fr";
end

import ("K1R4D3L.KillTracker.Locale."..locale);

_G.saveData = true;

_G.isLoaded = false;

hitLog = {}; -- timestamp, initiatorName, targetName
killLog = {}; -- timestamp, initiatorName, targetName, playerLevel
combatStateLog = {}; -- timestamp, inCombat
newKills = {}; -- new kill count

-- debugLevels:
--         0 -> quiet, only plugin load/unload and data load failure messages
--         1 -> basic debug info
--         2 ->   + all messages processed
--         3 ->   + hit/kill/in-combat table dumps with each hit/kill match attempt
_G.debugLevel = 0;

-- show log messages in bright yellow text?
_G.logHightlight = false;

-- send log messages > level 0 to script log instead of chat
_G.logToScriptLog = false;

-- grace period for hit-matching and in-combat checks (seconds)
delaySec = 2;

-- set minimum reasonable window dimensions
_G.minWidth = 520;
_G.minHeight = 200;

-- sort indicators - the unicode characters are nice, but only supported in Arial12
_G.DescSortChar = "+";
_G.AscSortChar = "-"; 
--_G.DescSortChar = string.char(226)..string.char(150)..string.char(188);     
--_G.AscSortChar = string.char(226)..string.char(150)..string.char(178);

_G.options = nil;

-- initialize the plugin
function Init()
   -- get plugin version
   local err, versionNo = pcall(GetVersionNo);
   _G.versionNo = (err and versionNo or "");

   -- get player
   _G.player = Turbine.Gameplay.LocalPlayer.GetInstance();
   player.name = player:GetName();

   -- load window size/position settings
   LoadSettings();
   
   -- create main KillTracker UI window, hidden
   CreateMainWindow();
   
   -- load server labels for same-named, cross-server character differentiation
   LoadServers();
	
   -- load any saved data
   LoadData();
   
   -- load unique server ID for same-named, cross-server character differentiation
   LoadServerID();
   
   InitCharacterNames();
   
   -- register shell command to display main window
   AddShellCommands();

   -- register for chat messages
   AddCallback(Turbine.Chat, "Received", ProcessChatMessages);

   -- register for in-combat state changes
   AddCallback(player, "InCombatChanged", ProcessCombatStateChange);

   -- initialize in-combat state
   InitCombatState();

	-- allow data processing to begin
	_G.isLoaded = true;
	
	-- log plugin startup complete
	LogMessage("KillTracker v"..versionNo..L.PluginLoaded..".", 0, false);
end

function InitCombatState()
	local inCombat = player:IsInCombat();
   local timestamp = Turbine.Engine.GetLocalTime();
      
   -- add current in-combat state to log table, with timestamp
   table.insert(combatStateLog, {timestamp, inCombat});

   LogMessage("KillTracker: InCombat="..tostring(inCombat)..","..timestamp, 2, false);
end

function ProcessCombatStateChange()
	if (isLoaded) then
		local inCombat = player:IsInCombat();
	   local timestamp = Turbine.Engine.GetLocalTime();
	      
	   -- add current in-combat state to log table, with timestamp
	   table.insert(combatStateLog, {timestamp, inCombat});
	
	   LogMessage("KillTracker: InCombat="..tostring(inCombat)..","..timestamp, 2, false);
	      
	   -- attempt to ID more kills based on new data
	   CheckForKills(timestamp);
	end
end

function ProcessChatMessages(sender, args)
	if (isLoaded) then
	   -- only process Combat and Death channels
		if ((args.ChatType ~= Turbine.ChatType.EnemyCombat) and (args.ChatType ~= Turbine.ChatType.PlayerCombat) and (args.ChatType ~= Turbine.ChatType.Death)) then
			return;
		end
		
		local timestamp = Turbine.Engine.GetLocalTime();
		
	   -- strip formatting from message and parse using localized parser 
		local updateType, initiatorName, targetName = L.Parse(string.gsub(string.gsub(args.Message, "<rgb=#......>(.*)</rgb>", "%1"), "^%s*(.-)%s*$", "%1"), args.ChatType);
	
		if (updateType == nil) then
	      return;
	   end
		
	   local tmpMsg = "";
	   
	   -- a mob hit or reflect, add to hit log
		if (updateType == 1) then
	      tmpMsg = initiatorName.." hit "..targetName;
	
	      table.insert(hitLog, {timestamp, initiatorName, targetName});
	   -- a mob kill
		elseif (updateType == 9) then
	   	-- grab the character's current level for the log
	   	local playerLevel = player:GetLevel();
	
	      -- direct kill by character, credit the character with the kill
	      if (initiatorName == "$PLAYER$") then
	
	         UpdateKillsTable(targetName, playerLevel, timestamp);
	         
	         LogMessage("KillTracker: "..player.name.." killed "..targetName..".", 1, false);
	      -- not a direct kill by character, add to kill log
	      else 
	         tmpMsg = initiatorName.." killed "..targetName;
	
	         table.insert(killLog, {timestamp, initiatorName, targetName, playerLevel});
	      end
		end
		
	   -- log logic result and original message
		LogMessage("KillTracker: "..tmpMsg..","..timestamp..","..args.Message, 2, false);
	
	   -- attempt to ID more kills based on new data
	   CheckForKills(timestamp);
	end
end

-- attempt to ID kills to be credited to character by comparing kill log with hit and in-combat logs
function CheckForKills(timestamp)
   local i = 1;

   -- dump table contents to log
	if (debugLevel > 2) then
      if (#killLog > 0) then
			for x = 1, #killLog do
         	LogMessage("KillTracker: kill:"..killLog[x][1]..","..killLog[x][2]..","..killLog[x][3], 3, false);
      	end
      end

      if (#combatStateLog > 0) then
			for x = 1, #combatStateLog do
         	LogMessage("KillTracker: inCombat:"..combatStateLog[x][1]..","..tostring(combatStateLog[x][2]), 3, false);
      	end
      end

      if (#hitLog > 0) then
			for x = 1, #hitLog do
         	LogMessage("KillTracker: hit:"..hitLog[x][1]..","..hitLog[x][2]..","..hitLog[x][3], 3, false);
      	end
      end
   end

   -- loop through kills
   while ((#killLog > 0) and (i <= #killLog)) do
      -- toss out kills that have expired without a match
      if (killLog[i][1] < (timestamp - delaySec)) then
         LogMessage("KillTracker: Found no match for \""..killLog[i][2].." killed "..killLog[i][3].."\".", 1, false);

         table.remove(killLog, i);
      else
         local j = 1;
         
         -- initialize "in-combat state during kill" to false
         local isInCombat = false;
         
         -- loop through in-combat state log
         while ((#combatStateLog > 0) and (j <= #combatStateLog)) do
            -- find "combat ended" logs
            if (not combatStateLog[j][2]) then
               -- "combat ended" logs that have expired
               if (combatStateLog[j][1] < (killLog[i][1] - delaySec)) then
                  -- toss out this entry and any earlier entries (lazy cleanup)
                  while (j > 0) do
                     table.remove(combatStateLog, j);

                     j = j - 1;
                  end
               
                  -- set "in-combat state during kill" to false 
                  isInCombat = false;
               -- end if a "combat ended" log is hit that is too close to or after kill
               --    (assumes that the player was in-combat unless obviously not)  
               else
                  j = #combatStateLog;
               end
            -- find "combat started" logs
            else
               -- set "in-combat state during kill" to true 
               isInCombat = true;
            end
            
            j = j + 1;
         end
            
         -- if the player was in combat during the kill, check for a matching hit
         if (isInCombat) then
            local k = 1;

            -- loop through hit log
            while ((#hitLog > 0) and (k <= #hitLog)) do
               -- toss out expired hits
               if (hitLog[k][1] < (killLog[i][1] - delaySec)) then
                  table.remove(hitLog,k);
               else
                  -- look for hit with same initiator and target as kill
                  if ((killLog[i][2] == hitLog[k][2]) and (killLog[i][2] == hitLog[k][2])) then
                     -- credit the character with the kill
                     UpdateKillsTable(killLog[i][3], killLog[i][4], killLog[i][1]);
                     
                     LogMessage("KillTracker: "..killLog[i][2].." killed "..killLog[i][3]..".", 1, false);

                     -- remove the matching rows from the logs and move to next kill
                     table.remove(killLog, i);
                     table.remove(hitLog, k);

                     break;
                  end

                  k = k + 1;
               end
            end

            i = i + 1;
         -- player wasn't in combat during kill, don't give credit
         elseif (#combatStateLog == 0) then
            LogMessage("KillTracker: Player not in combat during \""..killLog[i][2].." killed "..killLog[i][3].."\".", 1, false);

            -- toss out the uncredited kill
            table.remove(killLog, i);
         else
            i = i + 1;
         end
      end
   end
end

-- record newly-credited kills to master kill log 
function UpdateKillsTable(targetName, playerLevel, timestamp)
   local isTableUpdated = false;
   
   local i = 1;

   -- loop through master kill log
   while ((#killsTable > 0) and (i <= #killsTable)) do
      -- look for matching targets for this character, update stats and exit
      if ((killsTable[i][1] == targetName) and (killsTable[i][5] == player.uniqueName)) then
         -- increment kill count for this target
         killsTable[i][2] = killsTable[i][2] + 1;

         if (newKills[killsTable[i][1]] == nil) then
         	newKills[killsTable[i][1]] = 1;
         else
				newKills[killsTable[i][1]] = newKills[killsTable[i][1]] + 1;
			end  

         -- update character level at latest kill for this target
         killsTable[i][4] = playerLevel;

         -- update timestamp at latest kill for this target
         killsTable[i][6] = timestamp;

         isTableUpdated = true;

         break;
      end
      
      i = i + 1;   
   end
   
   -- need to add a new entry for this target for this character   
   if (not isTableUpdated) then
      -- log target name, initial kill count for this target of 1, character level
      --    at first kill of this target, character level at latest kill for this
      --    target, unique player ID (name + serverID), timestamp at latest kill for this target     
      table.insert(killsTable, {targetName, 1, playerLevel, playerLevel, player.uniqueName, timestamp});

		newKills[targetName] = 1;
   end
   
	if (debugLevel > 2) then
		for k, v in pairs(newKills) do
       	LogMessage("KillTracker: newKill:"..k..":"..v, 3, false);
    	end
	end

   -- save data to Account-level file
   SaveData();

   -- refresh the main KillTracker window
   RefreshKillsDisplay();
end

-- build character name lookup table for display, removing the ServerID part if
--    that character name is unique in the data, including the current player
function InitCharacterNames()
	_G.characterNames = {};

	local characterShortNames = {}; 

	for i = 1, #killsTable do
		-- if this is a new characterName + ServerID
		if (characterNames[killsTable[i][5]] == nil) then
			-- if the shortened name (just characterName, no Server ID) has not been used yet
			--    set the display name for the character to the shortened name
			if (characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)] == nil) then
				characterNames[killsTable[i][5]] = string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1);
			-- there are duplicate character names, switch them both to display the long names
			else
				characterNames[killsTable[i][5]] = killsTable[i][5];
				characterNames[characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)]] = characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)];
			end
				
			-- store the full name under the short name key in case we later find a duplicate
			characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)] = killsTable[i][5];
		end
	end
	
	-- add the current player if they're not already in the data
	if (characterNames[player.uniqueName] == nil) then
		if (characterShortNames[player.name] == nil) then
			characterNames[player.uniqueName] = player.name;
		else
			characterNames[player.uniqueName] = player.uniqueName;
			characterNames[characterShortNames[player.name]] = characterShortNames[player.name];
		end
	end
end

-- initialize plugin
Init();

-- options panel
plugin.GetOptionsPanel = function(self)
	if (options == nil) then
		_G.options = OptionsPanel();
	end

	return options;
end

-- register unload cleanup function
Plugins["KillTracker"].Unload = function(self, sender, args)
   RemoveShellCommands();

   -- remove chat message subscription
   RemoveCallback(Turbine.Chat, "Received", ProcessChatMessages)   
   -- remove in-combat state change subscription
   RemoveCallback(player, "InCombatChanged", ProcessCombatStateChange)   

   -- log plugin unload complete
   LogMessage("KillTracker v"..versionNo..L.PluginUnloaded..".", 0, false);
end
